/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capture.h>
#include <args.h>
#include <pcap.h>

#define READ_TIMEOUT 500
#define DEFAULT_MTU  16384

struct fpcap_priv {
	pcap_t		*pcap_desc;
	struct packet	pkt;
	struct proto	*proto;
	char		*ifname;
};

static int cb_if(struct arg *a, void *priv)
{
	struct fpcap_priv *p=priv;

	if ( !p ) return 0;

	if ( p->ifname ) {
		mesg(M_ERR,"pcap: Can't specify two %s entries!", a->name);
		return 0;
	}

	if ( !(p->ifname=strdup(a->val.v_str)) ) {
		return 0;
	}

	return 1;
}

static int finish_init(struct fpcap_priv *p, struct capture *cap, int live)
{
	unsigned int lnk;
	
	lnk = pcap_datalink(p->pcap_desc);

	p->proto = decode_subproto("__pcap_dlt", lnk);
	if ( p->proto == NULL ) {
		mesg(M_ERR,"pcap: %s: Can't support protocol 0x%x",
			p->ifname, lnk);
		return 0;
	}

	capdev_init_packet(&p->pkt, cap);
	return 1;
}

static int file_init(const char *args, struct capture *capture, void **priv)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	struct fpcap_priv *p;
	struct arg pl_args[]={
		{"file", ARGTYPE_STRING, cb_if},
		{NULL, ARGTYPE_NOP, NULL}
	};

	p = calloc(1, sizeof(*p));
	if ( p == NULL )
		return 0;

	switch( args_parse(pl_args, args, p) ) {
	case -1:
		mesg(M_ERR,"pcap: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	p->pcap_desc = pcap_open_offline(p->ifname, ebuf);
	if ( p->pcap_desc == NULL ) {
		mesg(M_ERR,"pcap: %s", ebuf);
		goto err;
	}

	if ( !finish_init(p, capture, 0) )
		goto err;

	*priv = p;
	return 1;
err:
	if ( p->ifname )
		free(p->ifname);
	free(p);
	return 0;
}

static int live_init(const char *args, struct capture *capture, void **priv)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	struct fpcap_priv *p;
	unsigned int mtu = 0, promisc = 0;
	struct arg pl_args[]={
		{"if", ARGTYPE_STRING, cb_if},
		{"mtu", ARGTYPE_PUINT, NULL, {.vp_uint = &mtu}},
		{"promisc", ARGTYPE_PBOOL, NULL, {.vp_bool = &promisc}},
		{NULL, ARGTYPE_NOP, NULL}
	};

	p = calloc(1, sizeof(*p));
	if ( p == NULL )
		return 0;

	switch( args_parse(pl_args, args, p) ) {
	case -1:
		mesg(M_ERR,"pcap: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	if ( !mtu )
		mtu = DEFAULT_MTU;

	ebuf[0] = '\0';
	p->pcap_desc = pcap_open_live(p->ifname, mtu, promisc,
					READ_TIMEOUT, ebuf);
	if ( p->pcap_desc == NULL ) {
		mesg(M_ERR,"pcap: %s", ebuf);
		goto err;
	}

	if ( ebuf[0] != '\0' )
		mesg(M_WARN, "pcap: %s\n", ebuf);

	if ( !finish_init(p, capture, 0) )
		goto err;

	*priv = p;
	return 1;
err:
	if ( p->ifname )
		free(p->ifname);
	free(p);
	return 0;
}
static void pcap_end(void *priv)
{
	struct fpcap_priv *p = priv;
	struct pcap_stat stat;

	if ( p == NULL )
		return;

        if (!pcap_file(p->pcap_desc) && pcap_stats(p->pcap_desc, &stat) < 0) {
		mesg(M_ERR, "pcap: pcap_stats(): %s",
			pcap_geterr(p->pcap_desc));
        }else{
		mesg(M_INFO,"pcap: received %u packets, dropped %u",
			stat.ps_recv, stat.ps_drop);
	}

	if ( p->pcap_desc )
		pcap_close(p->pcap_desc);

	if ( p->ifname )
		free(p->ifname);

	free(p);
}

static void lpf_callback(u_char *user, struct pcap_pkthdr *header, u_char *data)
{
	struct fpcap_priv *p;

	p = (struct fpcap_priv *)user;
	if ( p == NULL )
		return;

	/* Pointers */
	p->pkt.base = data;
	p->pkt.end = data;
	p->pkt.end += header->caplen;

	/* length */
	p->pkt.caplen = header->caplen;
	p->pkt.len = header->len;

	/* Timestamp */
	p->pkt.time.tv_sec = header->ts.tv_sec;
	p->pkt.time.tv_usec = header->ts.tv_usec;

	/* first layer */
	p->pkt.layer[0].proto = p->proto;
	p->pkt.layer[0].h.raw = p->pkt.base;
	p->pkt.layer[0].flags = 0;
	p->pkt.layer[0].session = NULL;
	p->pkt.llen = 0;

	p->pkt.alert = NULL;
	p->pkt.flags = FP_COPY|FP_REAL|FP_PROMISC|FP_WRITABLE;

	serial_number(&p->pkt);
	p->proto->decode(&p->pkt);
}

static void live_go(void *priv, struct capture *c)
{
	struct fpcap_priv *p = priv;
	int ret;
	
	while(c->state==CAP_STATE_CAPTURE) {
		ret = pcap_dispatch(p->pcap_desc, -1,
			(pcap_handler)lpf_callback, priv);

		if ( ret < 0 ) {
			mesg(M_ERR, "pcap: %s", pcap_geterr(p->pcap_desc));
			c->state=CAP_STATE_STOP;
			break;
		}
	}
}

static void file_go(void *priv, struct capture *c)
{
	struct fpcap_priv *p = priv;
	int ret;
	
	ret = pcap_dispatch(p->pcap_desc, 1,
		(pcap_handler)lpf_callback, priv);

	if ( ret < 0 )
		mesg(M_ERR, "pcap: %s", pcap_geterr(p->pcap_desc));

	if ( ret <= 0 )
		c->state = CAP_STATE_STOP;
}

struct capdev firestorm_capdevs[]={
	{
		.name = "pcap",
		.desc = "libpcap live capture",
		.init = live_init,
		.end = pcap_end,
		.go = live_go,
		.live = 1,
	},
	{
		.name = "pcapfile",
		.desc = "libpcap offline capture",
		.init = file_init,
		.end = pcap_end,
		.go = file_go,
		.live = 0,
	},
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capture.pcap",
	.desc		= "libpcap capture",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 0,
};
