/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * Capdev plugin which uses mmap to read libpcap files.
 *
 * TODO:
 *  o Sliding window to support large files.
*/
#include <firestorm.h>

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capture.h>
#include <args.h>

#ifndef lib_pcap_h

#define pcap_file_header tcpd_file_header
struct tcpd_file_header {
	unsigned int		magic;
	unsigned short		version_major;
	unsigned short		version_minor;
	int			thiszone;
	unsigned int		sigfigs;
	unsigned int		snaplen;
	unsigned int		linktype;
};

#define pcap_pkthdr tcpd_pkthdr
struct tcpd_pkthdr {
	struct timeval		ts;
	unsigned int		caplen;
	unsigned int		len;
};
#endif /* lib_pcap_h */

static struct {
	char *name;
	unsigned int magic;
	size_t size;
	int swap;
}magics[]={
	{"standard",			0xa1b2c3d4, 16, 0},
	{"redhat",			0xa1b2cd34, 24, 0},
	{"byte-swapped standard",	0xd4c3b2a1, 16, 1},
	{"byte-swapped redhat",		0x34cdb2a1, 24, 1},
	{NULL, 0, 0}
};

/* This is our own private data */
struct tcpd_priv {
	struct packet	pkt;
	u_int32_t	def_flags;
	void		*end;
	void		*cur;
	int		swap;
	size_t		phsiz;
	struct proto	*protocol;
	void		*map;
	unsigned int	map_size;
	int		fd;
	size_t		snaplen;
	unsigned int	(*r32)(unsigned int);
};

int cb_file(struct arg *, void *);
struct arg tcpd_args[]={
	{"file", ARGTYPE_STRING, cb_file},
	{NULL, ARGTYPE_NOP, NULL}
};

static u_int32_t read32(u_int32_t x){return x;}
static u_int32_t read32_swap(u_int32_t x){return bswap32(x);}

int cb_file(struct arg *a, void *priv)
{
	struct stat st;
	struct tcpd_priv *p = priv;
	struct pcap_file_header *fh;
	int i;

	if ( p == NULL )
		return 0;

	if ( p->fd > -1 ) {
		mesg(M_ERR,"tcpdump: Can't capture from more than one file");
		return 0;
	}

	/* Interface for us is just the filename  */
	if ( (p->fd=open(a->val.v_str, O_RDONLY)) < 0 ) {
		mesg(M_ERR,"tcpdump: %s: open(): %s",
			a->val.v_str, sys_err());
		return 0;
	}

	/* get the size of the file so we can map it in  */
	if ( fstat(p->fd, &st) ) {
		mesg(M_ERR,"tcpdump: %s: fstat(): %s",
			a->val.v_str, sys_err());
		close(p->fd);
		return 0;
	}

	p->map_size = (size_t)st.st_size;

	/* Check the file is actually even big enough */
	if ( p->map_size < sizeof(struct pcap_file_header) ) {
		mesg(M_ERR,"tcpdump: %s: Not a valid libpcap file",
			a->val.v_str);
		close(p->fd);
		return 0;
	}

	/* Memory map that sucker */
	if ( (p->map=mmap(NULL, p->map_size, PROT_READ,
		MAP_SHARED, p->fd, 0))==MAP_FAILED) {
		mesg(M_ERR,"tcpdump: %s: mmap(): %s",
			a->val.v_str, sys_err());
		close(p->fd);
		return 0;
	}
	p->end = p->map + p->map_size;

#if HAVE_MADVISE && defined(MADV_SEQUENTIAL)
	madvise(p->map, p->map_size, MADV_SEQUENTIAL);
#endif

	fh=(struct pcap_file_header *)p->map;

	/* Check what format the file is */
	for(p->phsiz=i=0; magics[i].name; i++) {
		if ( fh->magic == magics[i].magic ) {
			p->def_flags = FP_PROMISC|FP_REAL;

			if ( magics[i].swap ) {
				p->r32 = read32_swap;
				p->def_flags |= capdev_endian_flag(1);
			}else{
				p->r32 = read32;
				p->def_flags |= capdev_endian_flag(0);
			}

			p->phsiz = magics[i].size;
			p->snaplen = p->r32(fh->snaplen);
			mesg(M_INFO,"tcpdump: %s: %s: snaplen=%u",
				a->val.v_str, magics[i].name, p->snaplen);
			break;
		}
	}

	p->pkt.usage = 1;

	if ( !p->phsiz ) {
		mesg(M_ERR,"tcpdump: %s: Bad voodoo magic (0x%x)",
			a->val.v_str, fh->magic);
		munmap(p->map, p->map_size);
		close(p->fd);
		return 0;
	}

	/* Make sure we can decode this link type, not much point
	 * carrying on if we can't decode anything ;)  */
	p->protocol = decode_subproto("__pcap_dlt", p->r32(fh->linktype));
	if ( p->protocol == NULL ) {
		mesg(M_ERR,"tcpdump: %s: Unknown link type (0x%x)",
			a->val.v_str, p->r32(fh->linktype));
		munmap(p->map, p->map_size);
		close(p->fd);
		return 0;
	}

	return 1;
}

/* Initialise a capture process, we open the file and then
 * return our opaque private data structure to firestorm */
static int tcpd_init(const char *args, struct capture *capture, void **priv)
{
	struct tcpd_priv *p=NULL;

	if ( !(p=malloc(sizeof(*p))) )
		return 0;

	p->fd = -1;

	switch ( args_parse(tcpd_args, args, p) ) {
	case -1:
		mesg(M_ERR,"tcpdump: parse error: %s", args);
	case 0: /* fall through */
		free(p);
		return 0;
	default:
		break;
	}

	/* Initialise final few fields from header */
	p->cur = p->map;
	p->cur += sizeof(struct pcap_file_header);

	capdev_init_packet(&p->pkt, capture);

	*priv = p;
	return 1;
}

/* Stop the capture and cleanup our process */
static void tcpd_end(void *priv)
{
	struct tcpd_priv *p=priv;

	if ( !p ) return;

	if ( p->map ) munmap(p->map, p->map_size);
	if ( p->fd>=0 ) close(p->fd);

	free(p);
}

static void tcpd_go(void *priv, struct capture *c)
{
	struct pcap_pkthdr *h;
	struct tcpd_priv *p=priv;

	if ( !p )
		return;

	while(c->state == CAP_STATE_CAPTURE) {
		/* Make sure a packet header is present */
		if ( (p->cur + p->phsiz) > p->end ) {
			c->state=CAP_STATE_STOP;
			return;
		}

		/* Advance the p->cur to be the start of the
		 * actual packet data */
		h = (struct pcap_pkthdr *)p->cur;
		p->cur += p->phsiz;

		/* Check the packet is present */
		if ( (p->cur + p->r32(h->caplen)) > p->end ) {
			c->state = CAP_STATE_STOP;
			return;
		}

		/* Fill in the struct packet stuff */
		p->pkt.time.tv_sec = p->r32(h->ts.tv_sec);
		p->pkt.time.tv_usec = p->r32(h->ts.tv_usec);
		p->pkt.len = p->r32(h->len);
		p->pkt.caplen = p->r32(h->caplen);
		p->pkt.base = p->cur;
		p->pkt.end = p->cur+p->pkt.caplen;
		p->pkt.alert = NULL;
		p->pkt.flags = p->def_flags;

		/* advance the file pointer */
		p->cur += p->pkt.caplen;

		/* Dispatch to the firestorm decode engine */
		p->pkt.layer[0].proto = p->protocol;
		p->pkt.layer[0].h.raw = p->pkt.base;
		p->pkt.layer[0].flags = 0;
		p->pkt.layer[0].session = NULL;
		p->pkt.llen = 0;

		serial_number(&p->pkt);

		p->protocol->decode(&p->pkt);
	}
}

struct capdev firestorm_capdevs[]={
	{
		.name = "tcpdump",
		.desc = "A faster tcpdump file reader",
		.init = tcpd_init,
		.end = tcpd_end,
		.go = tcpd_go,
		.live = 0,
	},
	{ .name = NULL }
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capture.tcpdump",
	.desc		= "Reads packets in from tcpdump files",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 0,
};
