/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003 ECSC Ltd.
 * Copyright (c) 2004 Gianni Tedesco.
 * Author: Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
 *
 * Because different webservers require different normalization due to
 * non RFC-strict implementations and abiguities in the standards. This
 * file keeps track of which one the user has selected.
 *
 * TODO:
 *  o Need to add more emulators for other webservers
*/

#include <firestorm.h>
#include <errno.h>
#include <strtouint.h>
#include <string_hash.h>
#include "http_normalize.h"

#define IPHASH_SIZE 128

struct emulator {
	const char *name;
	unsigned long flags;
};

struct webserver {
	u_int32_t hash;
	u_int32_t ip;
	u_int16_t port;
	struct emulator *emu;
	struct webserver *next;
};

/* Emulation database */
static struct emulator emulators[]={
	{"IIS",		NADS_WS_UTF8 | NADS_WS_MS_UTF16 |
			NADS_WS_PERCENT_HACK | NADS_WS_DOUBLE_HEX |
			NADS_WS_BACKSLASHES },
	{"Apache",	NADS_WS_CASE_SENSITIVE | NADS_WS_UTF8 |
			NADS_WS_BACKSLASHES },
};

/* Default emulator */
static struct webserver default_ws = {
	.emu = &emulators[0],
};

/* IP:Port -> Webserver mapping */
static struct webserver *webservers[IPHASH_SIZE];

/* Currently selected webserver */
static struct webserver *current = &default_ws;

/* webserver_hash:
 * @host: Webserver host
 * @port: Webserver port
 *
 * Hash a unique identifier for a webserver
 *
 * Return value: 32bit hash
 */
static inline u_int32_t webserver_hash(u_int32_t host, u_int16_t port)
{
	u_int32_t h;

	h = host ^ port;
	h ^= port >> 8;
	h ^= h << 8;
	h ^= h << 12;

	return h;
}

/* find_emulator:
 * @name: Name of emulator to find
 *
 * Find an emulator based on it's name
 *
 * Return value: Pointer to the emulator, or NULL on error
 */
static struct emulator *find_emulator(const char *name)
{
	int num=sizeof(emulators)/sizeof(*emulators);
	int i;

	for(i=0; i<num; i++) {
		if ( !strcasecmp(emulators[i].name, name) ) {
			return &emulators[i];
		}
	}

	return NULL;
}

/* nads_default_webserver:
 * @name: Name of the webserver to protect
 *
 * Selects the webserver to use for any addresses we don't
 * know about.
 *
 * Return value: non-zero for success, zero for error.
 */
int nads_default_webserver(const char *name)
{
	struct emulator *e;

	e = find_emulator(name);
	if ( e == NULL )
		return 0;

	default_ws.emu = e;
	return 1;
}

/* nads_ws:
 * @x: the feature(s) we are checking for
 *
 * Returns true if ALL the requested features bits are set
 * by the currently selected webserver.
 *
 * Return value: 1 for true, 0 for false
 */
int __private nads_ws(unsigned long x)
{
	return ( (current->emu->flags & x) != 0 );
}

/* nads_webserver:
 * @ip: IP number of webserver
 * @port: TCP port it is listening on
 * @name: Name of the webserver to emulate
 *
 * Maps a webserver emulation type on to an ip:port tuple
 *
 * Return value: non-zero for success, zero for error.
 */
int nads_webserver(u_int32_t ip, u_int16_t port, const char *name)
{
	struct webserver *w;
	struct emulator *e;

	e = find_emulator(name);
	if ( e == NULL ) {
		errno = 0;
		return 0;
	}

	w = calloc(1, sizeof(*w));
	if ( w == NULL )
		return 0;

	w->emu = e;
	w->ip = ip;
	w->hash = webserver_hash(ip, port);
	w->port = port;
	w->next = webservers[w->hash % IPHASH_SIZE];
	webservers[w->hash % IPHASH_SIZE] = w;

	return 1;
}

/* select_webserver:
 * @host: Host name of server to select
 * @port: TCP port it is listening on
 *
 * Selects a webserver for comparisons, used from nads_normalize
 */
void __private select_webserver(u_int32_t ip, u_int16_t port)
{
	struct webserver *w;
	u_int32_t h;

	if ( ip == 0 || port == 0 ) {
		current = &default_ws;
		goto done;
	}

	if ( (current->port == port) &&
		current->ip == ip )
		goto done;

	h = webserver_hash(ip, port);

	for(w=webservers[h % IPHASH_SIZE]; w; w=w->next) {
		if ( (port == w->port) &&
			(ip == w->ip) ) {
			current = w;
			goto done;
		}
	}

	current = &default_ws;
done:
#if 0
	if ( current->ip )
		mesg(M_DEBUG, "Selected %s emulator for %s:%u",
			current->emu->name,
			iptostr(current->ip),
			current->port);
#endif
	return;
}
