/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include "tcpip.h"

static struct sub_table ip_subs[]={
	{"src", "ip.src"},
	{"dst", "ip.dst"},
	{"tos", "ip.tos"},
	{"id", "ip.id"},
	{"fragoffset", "ip.fragofs"},
	{"fragbits", "ip.fragbits"},
	{"ttl", "ip.ttl"},
	{"ip_proto", "ip.proto"},
	{"sameip", "ip.sameip"},
	{"ipopts", "ip.opts"},
	{"sport", "ip.sport"},
	{"dport", "ip.dport"},
	{NULL, NULL},
};

/* Decode Stuff */
struct proto ipv4_p={
	.name 		= "ip",
	.decode 	= ipv4_decode,
	.print 		= ipv4_dprint,
	.sig_add 	= ip_commit,
	.sig_match 	= ip_match,
	.sig_subs	= ip_subs,
	.fields 	= ip_fields,
	.tag		= ip_tag,
};

struct proto_req ipv4_r[]={
	proto_request("ethernet", __constant_htons(0x0800)),
	proto_request("sll", __constant_htons(0x0800)),
	proto_request("linux", __constant_htons(0x0800)),
	proto_request("__pcap_dlt", 12), /* DLT_RAW */
	proto_request("null", 0x02),
	null_request()
};

/* Token bucket for ip/ipfrag alerts */
struct tokenbucket ip_tb={
	.cost=(RATE_SEC/1), /* 1 alert per second */
	.burst=(RATE_SEC/1)*10, /* a burst of 10 alerts per second */
};

/* Generator */
struct generator ip_gen=init_generator("sig.ip", NULL);
struct generator dip_gen=init_generator("decode.ip", &ip_tb);

void ip_tag(struct packet *p, int l, struct alert_tag *t)
{
	if ( p->llen > l+1 ) {
		struct proto *n = p->layer[l+1].proto;
		if ( n == &udp_p || n == &tcp_p || n == &icmp_p ) {
			n->tag(p, ++l, t);
			return;
		}
	}

	mesg(M_DEBUG, "IP packet tagged");
}

/* Print out an IPv4 header */
int ipv4_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	struct in_addr s, d;
	char sip[16], dip[16];
	char end[32];
	struct pkt_iphdr *iph=p->layer[l].h.ip;

	s.s_addr=iph->saddr;
	d.s_addr=iph->daddr;
	snprintf(sip, sizeof(sip), "%s", inet_ntoa(s));
	snprintf(dip, sizeof(dip), "%s", inet_ntoa(d));

	if ( iph->frag_off & htons(IP_MF|IP_OFFMASK) ) {
		snprintf(end, sizeof(end), "(frag %u:%u@%u%s)",
		ntohs(iph->id),
		ntohs(iph->tot_len)-(iph->ihl<<2),
		(ntohs(iph->frag_off)&IP_OFFMASK)<<3,
		(ntohs(iph->frag_off)&IP_MF) ? "+" : "");
	}else if ( p->layer[l].flags & FLAG_IP_REASM ) {
		snprintf(end, sizeof(end),
			"(reasm %u:%u@0)",
			ntohs(iph->id),
			ntohs(iph->tot_len)-(iph->ihl<<2));
	}else{
		snprintf(end, sizeof(end),
			"len=%u", ntohs(iph->tot_len));
	}

	return snprintf(buf, buflen,
		"%s > %s ttl=%u proto=%u %s%s",
		sip,dip,iph->ttl,iph->protocol,end,
		(p->layer[l].flags & FLAG_IP_CSUM) ?
		"" : " !!bad_csum!!");
}

static inline u_int32_t ip_csum(struct pkt_iphdr *iph)
{
	u_int16_t *tmp=(u_int16_t *)iph;
	u_int16_t csum;
	u_int32_t sum=0;
	int i;

	for(i=0; i<iph->ihl<<1; i++) {
		sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	while(sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	csum=(~sum) & 0xffff;

	return csum ? 0 : FLAG_IP_CSUM;
}

/* Decode an IPv4 packet */
void ipv4_decode(struct packet *p)
{
	struct proto_child *pc;
	int my_layer=p->llen;
	struct layer *l=&p->layer[my_layer];
	int ks=0;

	/* Check it meets minimum size of packet */
	if ( l->h.raw + sizeof(struct pkt_iphdr) > p->end ) {
		goto nodecode;
	}

	/* Check to see if the whole datagram is here */
	if ( l->h.raw+ntohs(l->h.ip->tot_len) > p->end )
		l->flags|=FLAG_IP_DTRUNC;

	/* Check the header length isn't bogus, encapsulated
	 * protocols can't overlap the header! */
	if ( l->h.ip->ihl<5 ) {
		static struct alert a={
			.alert="Invalid IP header length",
			.sid=1, .rev=0,
			.priority=5,
			.gen=&dip_gen,
			.order = ALERT_BUILTIN,
		};
		alert_tag(p, &a, -1);
		goto nodecode;
	}

	/* We only accept IPv4 */
	if ( l->h.ip->version!=4 ) {
		static struct alert a={
			.alert="IP version != 4",
			.sid=2, .rev=0,
			.priority=5,
			.gen=&dip_gen,
			.order = ALERT_BUILTIN,
		};
		alert_tag(p, &a, -1);
		goto nodecode;
	}

	/* The total length can't be less than the header length. */
	if ( ntohs(l->h.ip->tot_len) < (l->h.ip->ihl<<2) ) {
		static struct alert a={
			.alert="IP datagram reported smaller than header",
			.sid=3, .rev=0,
			.priority=5,
			.gen=&dip_gen,
			.order = ALERT_BUILTIN,
		};
		alert_tag(p, &a, -1);
		goto nodecode;
	}

	/* We don't really need to check the checksum because you
	 * can't use bad IP header checksums to exploit anything, no
	 * routers will actually pass them on. Though you could
	 * generate attacks locally or it could be a genuine error.
	 * either way, its only 20 bytes! :) */
	l->flags|=ip_csum(l->h.ip);

	/* Use the ihl header field to decode the packet */
	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+(l->h.ip->ihl<<2))
		> p->end ) goto nodecode;

	/* We now know we have valid data */
	p->llen++;

	/* Perform IP de-fragmentation */
	if ( use_ipfrag && (l->h.ip->frag_off & ipfmask) ) {
		ks=ipfrag_process(p, my_layer);
	}

	/* Only decode fragments with offset 0, ie:
	 * those containing the next header */
	if ( l->h.ip->frag_off & __constant_htons(IP_OFFMASK) )
		goto nodecode;

	/* Don't decode too far */
	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		goto finish;
	}

	/* Find a relevent protocol handler to
	 * decode our encapsulated content */
	for(pc=l->proto->children; pc; pc=pc->next)
	{
		if ( l->h.ip->protocol == pc->id ) {
			p->layer[p->llen].flags=0;
			p->layer[p->llen].session=NULL;
			p->layer[p->llen].proto=pc->proto;
			pc->proto->decode(p);
			goto finish;
		}
	}

nodecode:
	if ( p->layer[p->llen].h.raw < p->end )
		p->layer[p->llen++].proto=NULL;
	dispatch(p);
finish:
	if ( ks ) ipq_kill(l->session);
}


/* ===================================================
 * PACKET MATCHING STUFF BEYOND THIS POINT
 * ===================================================
 */

static struct detect_root ip_root;

void ip_match(struct packet *p, int l)
{
	detect(&ip_root, p, l);
}

int ip_commit(struct rule *r)
{
	struct alert *a;
	struct sig_node *x;

	x = detect_new_sig(r, NULL);
	if ( x == NULL )
		return 0;

	a = calloc(1, sizeof(*a));
	if ( a == NULL ) {
		detect_free_sig(x, r->num_criteria);
		return 0;
	}

	memcpy(a, &r->alert, sizeof(*a));
	a->gen = &ip_gen;

	return detect_add_sig(&ip_root, x, r->num_criteria, a);
}
