/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/

#include "tcpip.h"
#include <netdb.h>

struct sip_detect {
	struct detect d;
	struct {
		struct alert *alert;
		struct detect *result;
	}sameip[2];
	struct detect *any;
};

static int get_ip_src(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohl(p->layer[l].h.ip->saddr);
	return 1;
}

static int get_ip_dst(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohl(p->layer[l].h.ip->daddr);
	return 1;
}

static int get_tcp_sport(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[++l].h.tcp->sport);
	return 1;
}

static int get_tcp_dport(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[++l].h.tcp->dport);
	return 1;
}

static int get_tcp_seq(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohl(p->layer[++l].h.tcp->seq);
	return 1;
}

static int get_tcp_ack(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohl(p->layer[++l].h.tcp->ack);
	return 1;
}

static int get_tcp_win(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohl(p->layer[++l].h.tcp->win);
	return 1;
}

static int get_tcp_flags(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[++l].h.tcp->flags.flags;
	return 1;
}

static int get_udp_sport(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[++l].h.udp->sport);
	return 1;
}

static int get_udp_dport(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[++l].h.udp->dport);
	return 1;
}

static int get_itype(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[++l].h.icmp->type;
	return 1;
}

static int get_icode(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[++l].h.icmp->code;
	return 1;
}

static int get_iid(struct packet *p, int l, idx_key_t *val)
{
	struct pkt_icmphdr *icmph=p->layer[++l].h.icmp;

	if ( icmph->type != ICMP_ECHO &&
			icmph->type != ICMP_ECHOREPLY )
		return 0;

	*val = ntohs(icmph->un.echo.id);
	return 1;
}

static int get_iseq(struct packet *p, int l, idx_key_t *val)
{
	struct pkt_icmphdr *icmph=p->layer[++l].h.icmp;

	if ( icmph->type != ICMP_ECHO &&
			icmph->type != ICMP_ECHOREPLY )
		return 0;

	*val = ntohs(icmph->un.echo.seq);
	return 1;
}

static int get_ip_ttl(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[l].h.ip->ttl;
	return 1;
}

static int get_ip_proto(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[l].h.ip->protocol;
	return 1;
}

static int parse_ip_proto(const char *args, idx_key_t *val)
{
	struct protoent *p;

	if ( !(p = getprotobyname(args)) )
		return 0;

	*val = p->p_proto;

	return 1;
}

static int get_ip_id(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[l].h.ip->id);
	return 1;
}

static int get_ip_ofs(struct packet *p, int l, idx_key_t *val)
{
	*val = ntohs(p->layer[l].h.ip->frag_off);
	*val ^= IP_OFFMASK;
	*val <<= 3;
	return 1;
}

static int get_ip_tos(struct packet *p, int l, idx_key_t *val)
{
	*val = p->layer[l].h.ip->tos;
	return 1;
}

static void sip_detect(struct detect *d, struct packet *pkt, int l)
{
	struct sip_detect *x = (struct sip_detect *)d;
	int i;

	i = (pkt->layer[l].h.ip->saddr == pkt->layer[l].h.ip->daddr);
	if ( x->sameip[i].alert )
		alert_tag(pkt, x->sameip[i].alert, l);
	if ( x->sameip[i].result )
		x->sameip[i].result->match(x->sameip[i].result, pkt, l);

	if ( x->any )
		x->any->match(x->any, pkt, l);
}

static struct detect **sip_default(struct detect *d)
{
	struct sip_detect *x = (struct sip_detect *)d;
	return &x->any;
}

static struct detect *sip_new(struct matcher *m)
{
	struct sip_detect *d;

	d = calloc(1, sizeof(*d));
	if ( d == NULL )
		return NULL;

	d->d.match = sip_detect;

	return (struct detect *)d;
}

static struct detect **sip_insert(struct detect *d, struct sig_node *n)
{
	struct sip_detect *x = (struct sip_detect *)d;
	int i = !n->negate;

	x->sameip[i].alert = alert_best(x->sameip[i].alert, n->alert);
	return &x->sameip[i].result;
}

static int sip_validate(struct matcher *match, char *args,
			void **priv, struct modifier *m)
{
	*priv=NULL;
	return 1;
}

static struct matcher sip_matcher = {
	.validate = sip_validate,
	.new_node = sip_new,
	.insert_node = sip_insert,
	.insert_default_node = sip_default,
};

struct field ip_fields[] = {
	{
		.name = "ip.src",
		.desc = "IP Source Address",
		.type = IDX_ORDERED_UINT32,
		.matcher = &ip_src_matcher,
		.get = {.val = get_ip_src},
		.cost = 10,
	},
	{
		.name = "ip.dst",
		.desc = "IP Destination Address",
		.type = IDX_ORDERED_UINT32,
		.matcher = &ip_dst_matcher,
		.get = {.val = get_ip_dst},
		.cost = 11,
	},
	{
		.name = "ip.fragbits",
		.desc = "IP Fragmentation Flags",
		.type = IDX_INVALID,
		.matcher = &ip_fragbits_matcher,
		.cost = MCOST_NET + 6,
	},
	{
		.name = "ip.opts",
		.desc = "IP Options",
		.type = IDX_INVALID,
		.matcher = &ip_opts_matcher,
		.cost = MCOST_NET + 20,
	},
	{
		.name = "ip.ttl",
		.desc = "IP Time-to-live",
		.type = IDX_ORDERED_UINT8,
		.get = {.val = get_ip_ttl},
		.cost = MCOST_NET + 88,
	},
	{
		.name = "ip.proto",
		.desc = "IP Protocol",
		.type = IDX_ORDERED_UINT8,
		.get = {.val = get_ip_proto},
		.parse = {.val = parse_ip_proto},
		.cost = MCOST_NET + 9,
	},
	{
		.name "ip.id",
		.desc = "IP Fragment ID",
		.type = IDX_ORDERED_UINT16,
		.get = {.val = get_ip_id},
		.cost = MCOST_NET + 4,
	},
	{
		.name "ip.ofs",
		.desc = "IP Fragment Offset",
		.type = IDX_ORDERED_UINT16,
		.get = {.val = get_ip_ofs},
		.cost = MCOST_NET + 4,
	},
	{
		.name "ip.tos",
		.desc = "IP Type Of Service",
		.type = IDX_ORDERED_UINT8,
		.get = {.val = get_ip_tos},
		.cost = MCOST_NET + 1,
	},
	{
		.name = "ip.sameip",
		.desc = "IP Source Address Matches Destination",
		.type = IDX_INVALID,
		.cost = MCOST_NET + 16,
		.matcher = &sip_matcher,
	},
	{.name = NULL}
};

struct field tcp_fields[] = {
	{
		.name = "tcp.sport",
		.desc = "TCP Source Port",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_tcp_sport},
		.cost = 13,
	},
	{
		.name = "tcp.dport",
		.desc = "TCP Destination Port",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_tcp_dport},
		.cost = 13,
	},
	{
		.name = "tcp.seq",
		.desc = "TCP Sequence number",
		.type = IDX_ORDERED_UINT32,
		.ofs = 1,
		.get = {.val = get_tcp_seq},
		.cost = MCOST_TRANS + 4,
	},
	{
		.name = "tcp.ack",
		.desc = "TCP Acknowlege Number",
		.type = IDX_ORDERED_UINT32,
		.ofs = 1,
		.get = {.val = get_tcp_ack},
		.cost = MCOST_TRANS + 14,
	},
	{
		.name = "tcp.win",
		.desc = "TCP Window Size",
		.type = IDX_ORDERED_UINT32,
		.ofs = 1,
		.get = {.val = get_tcp_win},
		.cost = MCOST_TRANS + 14,
	},
	{
		.name = "tcp.flow",
		.desc = "TCP State",
		.type = IDX_INVALID,
		.ofs = 1,
		.matcher = &tcp_flow_matcher,
		.cost = 9,
	},
	{
		.name = "tcp.stateless",
		.desc = "TCP Stateless Matching",
		.type = IDX_INVALID,
		.ofs = 1,
		.matcher = &tcp_stateless_matcher,
		.compatible = "flow",
		.cost = 9,
	},
	{
		.name = "tcp.flags",
		.desc = "TCP Flags",
		.type = IDX_ORDERED_UINT8,
		.ofs = 1,
		.get = {.val = get_tcp_flags},
		.matcher = &tcp_flags_matcher,
		.cost = MCOST_TRANS + 12,
	},
	{.name = NULL}

};

struct field udp_fields[] = {
	{
		.name = "udp.sport",
		.desc = "UDP Source Port",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_udp_sport},
		.cost = 13,
	},
	{
		.name = "udp.dport",
		.desc = "UDP Destination Port",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_udp_dport},
		.cost = 13,
	},
	{.name = NULL}

};

struct field icmp_fields[] = {
	{
		.name = "icmp.type",
		.desc = "ICMP Message type",
		.type = IDX_ORDERED_UINT8,
		.ofs = 1,
		.get = {.val = get_itype},
		.cost = MCOST_TRANS + 12,
	},
	{
		.name = "icmp.code",
		.desc = "ICMP Message Code",
		.type = IDX_ORDERED_UINT8,
		.ofs = 1,
		.get = {.val = get_icode},
		.cost = MCOST_TRANS + 13,
	},
	{
		.name = "icmp.id",
		.desc = "ICMP Echo ID",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_iid},
		.cost = MCOST_TRANS + 13,
	},
	{
		.name = "icmp.seq",
		.desc = "ICMP Echo Sequence Number",
		.type = IDX_ORDERED_UINT16,
		.ofs = 1,
		.get = {.val = get_iseq},
		.cost = MCOST_TRANS + 13,
	},
	{.name = NULL}
};
