/******************************************************************************/
/* feedforward neural network routines                                        */
/* (c) 2003 jamie twycross, milieu3@ukshells.co.uk                            */
/* released under gnu gpl version 2                                           */
/* hacked to use doubles from lwneuralnet by peter van rossum                 */
/******************************************************************************/

#include "network.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

/************/
/* defaults */
/************/

#define DEFAULT_MOMENTUM	0.0
#define DEFAULT_LEARNING_RATE	0.25
#define RANDOM_LARGE_WEIGHTS	1.0
#define RANDOM_SMALL_WEIGHTS	0.01

/******************/
/* initialization */
/******************/

network_t *net_randomize(network_t *net, const double range)
{
	register unsigned int l, nu, nl;

	for(l = 1; l < net->no_of_layers; l++) 
		for(nu = 0; nu < net->layer[l].no_of_neurons; nu++) 
			for(nl = 0; nl < net->layer[l - 1].no_of_neurons; nl++) 
				net->layer[l].neuron[nu].weight[nl] = \
					2.0 * range * ((double) rand() / \
					RAND_MAX - 0.5);

	return net;
}

network_t *net_reset_deltas(network_t *net)
{
	register unsigned int l, nu, nl;

	for(l = 1; l < net->no_of_layers; l++) 
		for(nu = 0; nu < net->layer[l].no_of_neurons; nu++) 
			for(nl = 0; nl < net->layer[l - 1].no_of_neurons; nl++) 
				net->layer[l].neuron[nu].delta[nl] = 0.0;

	return net;
}

/*********************/
/* memory management */
/*********************/

static int allocate_layer(layer_t *layer, const unsigned int no_of_neurons)
{
	layer->no_of_neurons = no_of_neurons;
	if(!(layer->neuron = calloc(no_of_neurons, sizeof(neuron_t))))
		return 1;
	
	return 0;
}

static int allocate_weights(layer_t *lower, layer_t *upper)
{
	unsigned int n;

	for(n = 0; n < upper->no_of_neurons; n++) {
		if(!(upper->neuron[n].weight = \
				calloc(lower->no_of_neurons, sizeof(double))))
			return 1;
		if(!(upper->neuron[n].delta = \
				calloc(lower->no_of_neurons, sizeof(double))))
			return 1;
	}
	
	return 0;
}

network_t *net_allocate_l(const unsigned int no_of_layers, int *arglist)
{
	unsigned int l;
	network_t *net;

	/* allocate memory for the network */
	if(!(net = malloc(sizeof(network_t))))
		return NULL;
	net->no_of_layers = no_of_layers;
	if(!(net->layer = calloc(no_of_layers, sizeof(layer_t))))
		return NULL;
	for(l = 0; l < no_of_layers; l++) 
		if(allocate_layer(&net->layer[l], arglist[l]))
			return NULL;
	for(l = 1; l < no_of_layers; l++)
		if(allocate_weights(&net->layer[l - 1], &net->layer[l]))
			return NULL;

	/* abbreviations for input and output layer */
	net->input_layer = &net->layer[0];
	net->output_layer = &net->layer[no_of_layers - 1];

	/* default values for network constants */
	net->momentum = DEFAULT_MOMENTUM;
	net->learning_rate = DEFAULT_LEARNING_RATE;

	/* initialize weights and deltas */
	net_randomize(net, RANDOM_LARGE_WEIGHTS);
	net_reset_deltas(net);

	return net;
}

network_t *net_allocate(const unsigned int no_of_layers, ...)
{
	unsigned int l;
	int *arglist;
	va_list args;
	network_t *net;

	if(!(arglist = calloc(no_of_layers, sizeof(int))))
		return NULL;
	va_start(args, no_of_layers);
	for(l = 0; l < no_of_layers; l++) 
		arglist[l] = va_arg(args, int);
	va_end(args);
	net = net_allocate_l(no_of_layers, arglist);
	free(arglist);

	return net;
}

void net_free(network_t *net)
{
	unsigned int l, n;

	for(l = 0; l < net->no_of_layers; l++) {
		if(l != 0) 
			for(n = 0; n < net->layer[l].no_of_neurons; n++) {
				free(net->layer[l].neuron[n].weight);
				free(net->layer[l].neuron[n].delta);
			}
		free(net->layer[l].neuron);
	}
	free(net->layer);
	free(net);
}

/**********/
/* access */
/**********/

network_t *net_set_momentum(network_t *net, const double momentum)
{
	net->momentum = momentum;
	return net;
}

double net_get_momentum(const network_t *net)
{
	return net->momentum;
}

network_t *net_set_learning_rate(network_t *net, const double learning_rate)
{
	net->learning_rate = learning_rate;
	return net;
}

double net_get_learning_rate(const network_t *net)
{
	return net->learning_rate;
}

unsigned int net_get_no_of_inputs(const network_t *net)
{
	return net->input_layer->no_of_neurons;
}

unsigned int net_get_no_of_outputs(const network_t *net)
{
	return net->output_layer->no_of_neurons;
}

unsigned int net_get_no_of_layers(const network_t *net)
{
	return net->no_of_layers;
}

double net_get_weight(const network_t *net, const unsigned int l, \
		const unsigned int nl, const unsigned int nu)
{
	return net->layer[l].neuron[nu].weight[nl];
}

/***********/
/* file io */
/***********/

int net_fprint(FILE *file, const network_t *net)
{
	unsigned int l, nu, nl;

	/* write network dimensions */
	fprintf(file, "%u\n", net->no_of_layers);
	for(l = 0; l < net->no_of_layers; l++) 
		fprintf(file, "%u\n", net->layer[l].no_of_neurons);

	/* write network constants */
	fprintf(file, "%f\n", net->momentum);
	fprintf(file, "%f\n", net->learning_rate);
	fprintf(file, "%f\n", net->global_error);

	/* write network weights */
	for(l = 1; l < net->no_of_layers; l++) 
		for(nu = 0; nu < net->layer[l].no_of_neurons; nu++) 
			for(nl = 0; nl < net->layer[l - 1].no_of_neurons; nl++) 
				fprintf(file, "%f\n", net->layer[l]. \
						neuron[nu].weight[nl]);

	return 0;
}

network_t *net_fscan(FILE *file)
{
	unsigned int no_of_layers, l, nu, nl;
	int *arglist;
	network_t *net;

	/* read network dimensions */
	fscanf(file, "%u", &no_of_layers);
	if(!(arglist = calloc(no_of_layers, sizeof(int))))
		return NULL;
	for(l = 0; l < no_of_layers; l++) 
    		fscanf(file, "%i", &arglist[l]);

	/* allocate memory for the network */
	if(!(net = net_allocate_l(no_of_layers, arglist)))
		return NULL;

	/* read network constants */
	fscanf(file, "%lf", &net->momentum);
	fscanf(file, "%lf", &net->learning_rate);
	fscanf(file, "%lf", &net->global_error);

	/* read network weights */
	for(l = 1; l < net->no_of_layers; l++) 
		for(nu = 0; nu < net->layer[l].no_of_neurons; nu++) 
			for(nl = 0; nl < net->layer[l - 1].no_of_neurons; nl++) 
				fscanf(file, "%lf", &net->layer[l].neuron[nu]. \
						weight[nl]);

	free(arglist);
	return net;
}

int net_print(const network_t *net)
{
	return net_fprint(stdout, net);
}

int net_save(const network_t *net, const char *filename)
{
	FILE *file;

	if(!(file = fopen(filename, "w")))
		return(1);
	net_fprint(file, net);

	return fclose(file);
}

network_t *net_load(const char *filename)
{
	FILE *file;
	network_t *net;

	if(!(file = fopen(filename, "r")))
		return NULL;
	net = net_fscan(file);
	fclose(file);

	return net;
}

/********************/
/* input and output */
/********************/

static inline void set_input(network_t *net, const double *input)
{
	register unsigned int n;

	for(n = 0; n < net->input_layer->no_of_neurons; n++)
		net->input_layer->neuron[n].output = input[n];
}

static inline void get_output(const network_t *net, double *output)
{
	register unsigned int n;

	for(n = 0; n < net->output_layer->no_of_neurons; n++) 
		output[n] = net->output_layer->neuron[n].output;
}

/**********************/
/* sigmoidal function */
/**********************/

static inline double sigma(const double x)
{
	return (1.0 / (1.0 + exp(-1.0 * x)));
}

/************************************/
/* forward and backward propagation */
/************************************/

static inline void propagate_layer(layer_t *lower, layer_t *upper)
{
	register unsigned int nu, nl;
	register double value;

	for(nu = 0; nu < upper->no_of_neurons; nu++) {
		value = 0.0;
		for(nl = 0; nl < lower->no_of_neurons; nl++) 
			value += upper->neuron[nu].weight[nl] * \
				lower->neuron[nl].output;
		upper->neuron[nu].output = sigma(value);
	}
}

static inline void forward_pass(network_t *net)
{
	register unsigned int l;

	for(l = 1; l < net->no_of_layers; l++)
		propagate_layer(&net->layer[l - 1], &net->layer[l]);
}

double net_compute_output_error(network_t *net, const double *target)
{
	register unsigned int n;
	register double output, error;

	net->global_error = 0.0;
	for(n = 0; n < net->output_layer->no_of_neurons; n++) {
		output = net->output_layer->neuron[n].output;
		error = target[n] - output;
		net->output_layer->neuron[n].error = output * (1.0 - output) * error;
		net->global_error += error * error;
	}
	net->global_error *= 0.5;

	return net->global_error;
}

double net_get_output_error(const network_t *net)
{
	return net->global_error;
}

static inline void backpropagate_layer(layer_t *lower, layer_t *upper)
{
	register unsigned int nl, nu;
	register double output, error;

	for(nl = 0; nl < lower->no_of_neurons; nl++) {
		error = 0.0;
		for(nu = 0; nu < upper->no_of_neurons; nu++) 
			error += upper->neuron[nu].weight[nl] * \
				upper->neuron[nu].error;
		output = lower->neuron[nl].output;
		lower->neuron[nl].error = output * (1.0 - output) * error;
	}
}

static inline void backward_pass(network_t *net)
{
	register unsigned int l;

	for(l = net->no_of_layers - 1; l > 1; l--)
		backpropagate_layer(&net->layer[l - 1], &net->layer[l]);
}

static inline void adjust_weights(network_t *net)
{
	register unsigned int l, nu, nl;
	register double error, delta;

	for(l = 1; l < net->no_of_layers; l++)
		for(nu = 0; nu < net->layer[l].no_of_neurons; nu++) {
			error = net->layer[l].neuron[nu].error;
			for(nl = 0; nl < net->layer[l-1].no_of_neurons; nl++) {
				delta =	net->learning_rate * error * \
					net->layer[l - 1].neuron[nl].output +
					net->momentum * \
					net->layer[l].neuron[nu].delta[nl];
				net->layer[l].neuron[nu].weight[nl] += delta;
				net->layer[l].neuron[nu].delta[nl] = delta;
			}
		}
}

/***************************/
/* evaluation and training */
/***************************/

network_t *net_compute(network_t *net, const double *input, double *output)
{
	set_input(net, input);
	forward_pass(net);
	if(output != NULL)
		get_output(net, output);

	return net;
}

network_t *net_train(network_t *net)
{
	backward_pass(net);
	adjust_weights(net);

	return net;
}
