/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * This code handles alerts, it mainly just calls functions in
 * elog_write.c.
 *
 * Functions:
 *  o alert_ratelimit() - token bucket filter
 *  o alert_do_ratelimit() - wrapper to do the correct ratelimit
 *  o alert() - exported API, do an alert
*/
#include <firestorm.h>

#include <signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>

#include <alert.h>
#include <elog.h>
#include <elog_write.h>
#include <signature.h>
#include <decode.h>
#include <args.h>
#include <capture.h>
#include <target.h>

/* Token bucket rate-limiting filter. Return values:
 * returns -1 if events are coming in too fast
 * returns zero if rate is normal
 * returns n>0 if rate is normal but some were dropped */
static inline int alert_ratelimit(struct packet *pkt, struct tokenbucket *a)
{
	unsigned long h;
	int ret;

	if ( !a->cost )
		return 0;

	/* Build a single time value with 0.01s resolution */
	h=pkt->time.tv_sec * RATE_SEC;
	h+=pkt->time.tv_usec / (1000000UL/RATE_SEC);

	a->toks += h - a->last_msg;
	a->last_msg=h;

	if ( a->toks > a->burst )
		a->toks=a->burst;

	if ( a->toks >= a->cost ) {
		ret=a->missed;
		a->missed=0;
		a->toks -= a->cost;
		return ret;
	}

	a->missed++;
	return -1;
}

/* return zero for OK, -1 to suppress */
static int alert_do_ratelimit(struct tokenbucket *tb,
				struct packet *pkt,
				const char *s1, const char *s2)
{
	int missed;

	if ( !tb )
		return 0;

	switch ( (missed=alert_ratelimit(pkt,tb)) ) {
	case -1:
		return -1;
	case 0:
		return 0;
	default:
		mesg(M_DEBUG, "alert: [%s] %s: %i alerts suppressed",
			s1, s2, missed);
		return 0;
	}
}

void alert(struct packet *pkt)
{
	struct alert *a;

	if ( !pkt )
		return;
	if ( !pkt->alert )
		return;

	a = pkt->alert;

	notify_run(NOTIFY_PRE_ALERT, pkt, -1);

	/* Token bucket filters are hierarchical */
	if ( alert_do_ratelimit(&a->t, pkt, "sig", a->alert)<0 )
		return;

	if ( alert_do_ratelimit(a->gen->t, pkt, "generator",
				a->gen->name)<0 )
		return;

	notify_run(NOTIFY_ALERT, pkt, -1);
}
