/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is an argument parsing library designed to work on const strings,
* it is typesafe and presents a simple API. Supported types are strings,
* integers (signed and unsigned), double precision floats and boolean.
*/
#include <firestorm.h>
#include <dotscara/sys.h>
#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>
#include <dotscara/fibuf.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <errno.h>
#include <args.h>

/* first char is blank because arg_q[0] means no quotes */
static char arg_q[]=" \'\"";

/* Dispatch functions */
static int arg_string(struct arg *, char *, void *);
static int arg_bytes(struct arg *, char *, void *);
static int arg_off(struct arg *, char *, void *);
static int arg_uint(struct arg *, char *, void *);
static int arg_int(struct arg *, char *, void *);
static int arg_bool(struct arg *, char *, void *);
static int arg_double(struct arg *, char *, void *);
static int arg_files(struct arg *, char *, void *);

/* Dispatch table */
typedef int(*proc_argdispatch)(struct arg *, char *, void *);
static proc_argdispatch arg_dfn[ARGTYPE_MAX]={
	[ ARGTYPE_NOP ] arg_string,
	[ ARGTYPE_UINT ] arg_uint,
	[ ARGTYPE_PUINT ] arg_uint,
	[ ARGTYPE_INT ] arg_int,
	[ ARGTYPE_PINT ] arg_int,
	[ ARGTYPE_BYTES ] arg_bytes,
	[ ARGTYPE_PBYTES ] arg_bytes,
	[ ARGTYPE_STRING ] arg_string,
	[ ARGTYPE_BOOL ] arg_bool,
	[ ARGTYPE_PBOOL ] arg_bool,
	[ ARGTYPE_DOUBLE ] arg_double,
	[ ARGTYPE_PDOUBLE ] arg_double,
	[ ARGTYPE_OFF ] arg_off,
	[ ARGTYPE_POFF ] arg_off,
	[ ARGTYPE_FIBUF ] arg_files,
	[ ARGTYPE_FOBUF ] arg_files,
};

/* Boolean data type */
static int arg_bool(struct arg *a, char *v, void *user)
{
	unsigned int r;

	if ( v == NULL ) {
		r=1;
	}else if ( !strcasecmp("true", v) ||
		!strcasecmp("yes", v) ||
		!strcasecmp("on", v) ) {
		r=1;
	}else if ( !strcasecmp("false", v) ||
		!strcasecmp("no", v) ||
		!strcasecmp("off", v) ) {
		r=0;
	}else
		return -1;

	if ( a->type==ARGTYPE_PBOOL ) {
		if ( a->val.vp_bool ) {
			*a->val.vp_bool=r;
			return 1;
		}else
			return -1;
	}else{
		if ( !a->fn )
			return -1;
		a->val.v_bool=r;
		return a->fn(a, user);
	}
}

/* Numeric datatype for representing bytes (eg: 1024k, 10M ) */
static int arg_bytes(struct arg *a, char *v, void *user)
{
	unsigned long int val;
	char *end;

	if ( v == NULL )
		return -1;

	val = strtoul(v, &end, 0);
	if ( end == v )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	switch( *end ) {
	case '\0':
		break;
	case 'k':
	case 'K':
		val <<= 10;
		break;
	case 'm':
	case 'M':
		val <<= 20;
		break;
	case 'g':
	case 'G':
		val <<= 30;
		break;
#if 0
	case 't':
	case 'T':
		val <<= 40;
		break;
	case 'p':
	case 'P':
		val <<= 50;
		break;
	case 'e':
	case 'E':
		val <<= 60;
		break;
	case 'z':
	case 'Z':
		val <<= 70;
		break;
	case 'y':
	case 'Y':
		val <<= 80;
		break;
#endif
	}

	if ( val > UINT_MAX )
		return -1;

	if ( a->type == ARGTYPE_PBYTES ) {
		if ( a->val.vp_bytes == NULL )
			return -1;

		*a->val.vp_bytes = val;
		return 1;
	}else{
		if ( !a->fn )
			return -1;

		a->val.v_bytes = val;
		return a->fn(a, user);
	}
}

/* Numeric datatype for representing off_t (eg: 1024k, 10M ) */
static int arg_off(struct arg *a, char *v, void *user)
{
#if _FILE_OFFSET_BITS == 64
	unsigned long long int val;
#else
	unsigned long int val;
#endif
	char *end;

	if ( v == NULL )
		return -1;

#if _FILE_OFFSET_BITS == 64
	val = strtoull(v, &end, 0);
#else
	val = strtoul(v, &end, 0);
#endif
	if ( end == v )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	switch( *end ) {
	case '\0':
		break;
	case 'k':
	case 'K':
		val <<= 10;
		break;
	case 'm':
	case 'M':
		val <<= 20;
		break;
	case 'g':
	case 'G':
		val <<= 30;
		break;
	case 't':
	case 'T':
		val <<= 40;
		break;
	case 'p':
	case 'P':
		val <<= 50;
		break;
	case 'e':
	case 'E':
		val <<= 60;
		break;
#if 0
	case 'z':
	case 'Z':
		val <<= 70;
		break;
	case 'y':
	case 'Y':
		val <<= 80;
		break;
#endif
	}

	if ( val > UINT_MAX )
		return -1;

	if ( a->type == ARGTYPE_POFF ) {
		if ( a->val.vp_off == NULL )
			return -1;

		*a->val.vp_off = val;
		return 1;
	}else{
		if ( !a->fn )
			return -1;

		a->val.v_off = val;
		return a->fn(a, user);
	}
}

/* Unsigned integer data type */
static int arg_uint(struct arg *a, char *v, void *user)
{
	unsigned long int val;
	char *end;

	if ( v == NULL )
		return -1;

	val = strtoul(v, &end, 0);
	if ( end == v )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	if ( val > UINT_MAX )
		return -1;

	if ( a->type == ARGTYPE_PUINT ) {
		if ( a->val.vp_uint == NULL )
			return -1;

		*a->val.vp_uint = val;
		return 1;
	}else{
		if ( !a->fn )
			return -1;

		a->val.v_uint = val;
		return a->fn(a, user);
	}
}

/* Signed integer data type */
static int arg_int(struct arg *a, char *v, void *user)
{
	unsigned long int val;
	char *end;
	int final = 0;

	if ( v == NULL )
		return -1;

	/* negative */
	if ( *v == '-' )
		final = 1;

	val = strtoul(v, &end, 0);
	if ( end == v )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	if ( val > (UINT_MAX>>1) )
		return -1;

	if ( final )
		final = -val;
	else
		final = val;

	if ( a->type == ARGTYPE_PINT ) {
		if ( a->val.vp_int == NULL )
			return -1;

		*a->val.vp_int = val;
		return 1;
	}else{
		if ( !a->fn )
			return -1;

		a->val.v_int = val;
		return a->fn(a, user);
	}
}

/* Double data type */
static int arg_double(struct arg *a, char *v, void *user)
{
	double val;
	char *end;

	if ( v == NULL )
		return -1;

	val = strtod(v, &end);
	if ( end == v )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == HUGE_VAL && errno == ERANGE )
		return -1;

	if ( a->type == ARGTYPE_PDOUBLE ) {
		if ( a->val.vp_double == NULL )
			return -1;

		*a->val.vp_double = val;
		return 1;
	}else{
		if ( !a->fn )
			return -1;

		a->val.v_double = val;
		return a->fn(a, user);
	}
}

/* String data type */
static int arg_string(struct arg *a, char *v, void *user)
{
	/* By specifying string or literal you
	 * are saying you want a string */
	if (a->type!=ARGTYPE_NOP && v==NULL)
		return -1;

	if ( !a->fn )
		return -1;

	a->val.v_str=v;

	return a->fn(a, user);
}

/* Files */
static int arg_files(struct arg *a, char *v, void *user)
{
	int flags = 0;
	int ret = 0;
	int fd;

	if ( a->val.fibuf == NULL )
		return -1;

	switch ( a->type ) {
	case ARGTYPE_FIBUF:
		flags = O_RDONLY;
		break;
	case ARGTYPE_FOBUF:
		flags = O_WRONLY|O_TRUNC|O_CREAT;
		break;
	default:
		return -1;
	}

	fd = open(v, flags, 0640);
	if ( fd < 0 ) {
		mesg(M_ERR, "args: %s: open(): %s", v, sys_err());
		return 0;
	}

	switch ( a->type ) {
	case ARGTYPE_FIBUF:
		ret = fibuf_attach(a->val.fibuf, fd, DSE_NATIVE_ENDIAN, 0);
		break;
	case ARGTYPE_FOBUF:
		ret = fobuf_attach(a->val.fobuf, fd, DSE_NATIVE_ENDIAN, 0);
		break;
	}

	if ( ret == 0 ) {
		mesg(M_ERR, "args: %s: fobuf_attach(): %s", v, sys_err2(NULL));
		__fd_close(fd);
		return 0;
	}

	return 1;
}

/* Copy and NULL terminate from a read-only vector
 * in to a pre-allocated read/write buffer. */
static size_t arg_strcpy(char *dst, size_t dlen, const char *src, size_t slen)
{
	char *dend;
	const char *send;
	int esc=0;
	size_t len=0;

	if ( !dst || !dlen || dlen+1 == 0) return 0;
	if ( !src || !slen ) {
		*dst=0;
		return 0;
	}

	dend=dst+dlen-1;
	send=src+slen;

	while ( dst<dend && src<send ) {
		if ( !esc && *(src)=='\\' ) {
			esc=1;
			src++;
		}else{
			*(dst++)=*(src++);
			esc=0;
			len++;
		}
	}

	*dst=0;
	return len;
}

/* Dispatch a key/value pair to the right place */
static int arg_found(struct arg *a,
	const char *key, size_t klen,
	const char *val, size_t vlen,
	void *user) {
	char kbuf[AKEY_SIZE];
	char vbuf[AVAL_SIZE];
	char *r;
	struct arg *tmp;
	char *sval;
	int ret=-1;

	if ( arg_strcpy(kbuf, sizeof(kbuf), key, klen) < klen )
		return -1;

	if ( arg_strcpy(vbuf, sizeof(vbuf), val, vlen) < vlen )
		return -1;

	if ( vbuf[0] == '\0' ) {
		sval=NULL;
	}else{
		sval=vbuf;
	}

	for(tmp=a; ; tmp++) {

		/* if name is passed but isn't right, keep looking */
		if ( tmp->name && strcmp(kbuf, tmp->name) )
			continue;

		/* Check it is possible to call back */
		if ( a->type >= ARGTYPE_MAX
			|| !arg_dfn[a->type] ) {
			ret=-1;
			break;
		}

		if ( !tmp->name ) {
			/* Fill in what the arg was so
			 * handler can do own matching */
			tmp->name=kbuf;
			r=NULL;
		}else{
			r=tmp->name;
		}

		ret=arg_dfn[tmp->type](tmp, sval, user);
		tmp->name=r;

		/* can only match once */
		break;
	}

	return ret;
}

int arg_dispatch(struct arg *a, const char *key, const char *val, void *priv)
{
	return arg_found(a, key, strlen(key),
		val, val ? strlen(val) : 0, priv);
}

/* Determine what kind of quote character this is */
static inline int arg_quote(char q)
{
	int i=0;

	for(i=1; arg_q[i]; i++) {
		if ( arg_q[i] == q )
			return i;
	}

	i=0;
	return i;
}

/* Actual API entry */
int args_parse(struct arg *a, const char *str, void *user)
{
	int ret=0;

	/* state machine tape position */
	const char *cur;

	/* Initial state */
	int state=0;
	int end=0;
	int inq=0;
	int esc=0;

	/* Key and value vectors */
	const char *key=NULL; size_t klen=0;
	const char *val=NULL; size_t vlen=0;

	if ( str == NULL )
		return 1;
	if ( a == NULL )
		return -1;

	for(cur=str; ; cur++) {
		if ( *cur==0 ) end=1;

		if ( state==0 ) { /* leading isspace */
			if ( !isspace(*cur) ) {
				state=1;
				key=cur;
				klen=1;
			}
		}else if ( state==1 ) { /* Inside a key */
			if ( *cur=='=' ) {
				val=cur+1;
				vlen=0;
				state=2;
			}else if ( isspace(*cur) || *cur==0 ) {
				if ( (ret=arg_found(a, key, klen,
					val, vlen, user))<=0 )
					return ret;
				state=0;
			}else klen++;
		}else if ( state==2 ) { /* Inside a value */
		
			/* Handle escaped characters */
			if ( esc > 2 ) esc=0;
			if ( !esc && *cur=='\\' ) {
				esc=1;
			}else if ( esc ){
				esc++;
			}

			if ( !inq ) {
				if ( !vlen && (inq=arg_quote(*cur)) ) {
					/* The first char is a quote char */
					val++;
				}else if ( isspace(*cur) || *cur=='\0') {
					/* No quotes */
					if ( !vlen )
						val=NULL;
					if ( (ret=arg_found(a, key, klen,
						val, vlen, user))<=0 )
						return ret;
					key=val=NULL;
					esc=state=inq=klen=vlen=0;
				}else vlen++;
			}else{
				/* Check for an end quote */
				if ( (!esc && *cur==arg_q[inq]) ||
					*cur==0) {
					if ( !vlen ) val=NULL;
					if ( (ret=arg_found(a, key, klen,
						val, vlen, user))<=0 )
						return ret;
					key=val=NULL;
					esc=state=inq=klen=vlen=0;
				}else vlen++;
			}
		}

		if ( end ) break;
	}

	return ret;
}

const char *args_typename(unsigned int type)
{
	switch ( type ) {
	case ARGTYPE_NOP:
		return "custom";
	case ARGTYPE_UINT:
	case ARGTYPE_PUINT:
		return "unsigned integer";
	case ARGTYPE_INT:
	case ARGTYPE_PINT:
		return "signed integer";
	case ARGTYPE_BYTES:
	case ARGTYPE_PBYTES:
		return "bytes";
	case ARGTYPE_STRING:
		return "string";
	case ARGTYPE_BOOL:
	case ARGTYPE_PBOOL:
		return "bool";
	case ARGTYPE_DOUBLE:
	case ARGTYPE_PDOUBLE:
		return "double";
	case ARGTYPE_OFF:
	case ARGTYPE_POFF:
		switch ( sizeof(off_t) ) {
		case 4:
			return "32bit file offset";
		case 8:
			return "64bit file offset";
		default:
			return "file offset";
		}
	case ARGTYPE_FIBUF:
		return "input file";
	case ARGTYPE_FOBUF:
		return "output file";
	default:
		return "unknown";
	}
}
