/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 *  o capture_cleanup() - stop the packet capture
 *  o capture_init() - register cleanup
 *  o capture_stop() - finish the current capture
 *  o capture_interrupt() - interrupt the current capture
 *  o capture_setup() - initialise the capture
 *  o capture_timestamp() - get a (cpu) timestamp
 *  o capture_go() - packet capture main loop
*/
#include <firestorm.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#include <unistd.h>

#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capture.h>

static void init_competing_signals(void);

static struct capture capture;

int capture_is_live(void)
{
	return !!capture.capdev->live;
}

/* Make sure to call the capdev cleanup function before we quit */
static void capture_cleanup(struct packet *pkt, int l, void *priv)
{
	if ( capture.priv )
		capture.capdev->end(capture.priv);
	if ( capture.args )
		free((char *)capture.args);
}

void __private capture_init(void)
{
	init_competing_signals();
	notify_exit(capture_cleanup, NULL);
}

void capture_interrupt(int type)
{
	if ( type == CAP_INTR_DEADLY ) {
		capture.state = CAP_STATE_STOP;
		return;
	}

	if ( capture.state == CAP_STATE_STOP )
		return;

	/* This is safe because we unset it with signals masked */
	capture.state = CAP_STATE_INTR;

	/* This is safe because competing signals are blocked
	 * (provided host has sigaction.
	 */
	capture.intr |= type;
}

/* Setup the capture structures to get going */
void capture_setup(const char *type, const char *args)
{
	struct capdev *cd;
	const char *a = args;

	if ( !type ) {
		firestorm_exit(EXIT_FAILURE, "capture: wtf?!");
	}

	if ( !(cd=capdev_find(type)) ) {
		firestorm_exit(EXIT_FAILURE, "capture: cant find '%s'", type);
	}

	if ( args ) {
		a = strdup(args);
		if ( a == NULL )
			firestorm_die_error("strdup");
	}

	capture.capdev = cd;
	capture.args = a;

	if ( !capture.capdev->init(args, &capture, &capture.priv) )
		firestorm_exit(EXIT_FAILURE, "capture: failed to initialise");

	capture.state=CAP_STATE_INIT;
}

/* Get a timestamp */
static inline void capture_timestamp(struct timeval *tv)
{
#ifdef HAVE_GETRUSAGE
	static struct rusage r;
	getrusage(RUSAGE_SELF, &r);
	tv->tv_sec=r.ru_utime.tv_sec;
	tv->tv_usec=r.ru_utime.tv_usec;
#else
	gettimeofday(tv, NULL);
#endif
}

#ifdef HAVE_SIGACTION
static sigset_t competing_signals;
static void init_competing_signals(void)
{
	sigemptyset(&competing_signals);
	sigaddset(&competing_signals, SIGINT);
	sigaddset(&competing_signals, SIGTERM);
	sigaddset(&competing_signals, SIGUSR1);
	sigaddset(&competing_signals, SIGUSR2);
	sigaddset(&competing_signals, SIGALRM);
	sigaddset(&competing_signals, SIGHUP);
}

static inline void mask_competing_signals(void)
{
	sigprocmask(SIG_BLOCK, &competing_signals, NULL);
}

static inline void unmask_competing_signals(void)
{
	sigprocmask(SIG_UNBLOCK, &competing_signals, NULL);
}
#else
static void init_competing_signals(void)
{
}
static inline void mask_competing_signals(void)
{
}
static inline void unmask_competing_signals(void)
{
}
#endif

static void dispatch_intrs(sig_atomic_t intrs)
{
	if ( intrs & CAP_INTR_HUP )
		notify_run(NOTIFY_HUP, NULL, -1);
	if ( intrs & CAP_INTR_TIMER )
		notify_run(NOTIFY_TIMER, NULL, -1);
	if ( intrs & CAP_INTR_USR1 )
		notify_run(NOTIFY_USR1, NULL, -1);
	if ( intrs & CAP_INTR_USR2 )
		notify_run(NOTIFY_USR2, NULL, -1);
}

/* Go in to the packet capture main loop */
void capture_go(void)
{
	struct timeval tv, tv2;
	struct timeval end;

	if ( capture.state!=CAP_STATE_INIT ) {
		firestorm_exit(EXIT_FAILURE, "No captures specified");
	}

	/* Ready to go */
	mesg(M_INFO,"capture: %s[%s]: started",
		capture.capdev->name, capture.args);

	/* Do the capture */
	capture.state=CAP_STATE_CAPTURE;
#ifdef PROFILE_MODE
	pause();
#endif
	capture_timestamp(&tv);
	do {
		capture.capdev->go(capture.priv, &capture);
		if ( capture.state == CAP_STATE_INTR ) {
			sig_atomic_t intrs;

			mask_competing_signals();
			intrs = capture.intr;
			capture.state = CAP_STATE_CAPTURE;
			capture.intr = 0;
			unmask_competing_signals();

			dispatch_intrs(intrs);
		}
	} while ( capture.state==CAP_STATE_CAPTURE );
	capture_timestamp(&tv2);
	capture.state=CAP_STATE_ENDED;

	/* All done */
	mesg(M_INFO,"capture: %s[%s]: stopped",
		capture.capdev->name,capture.args);

	/* Calculate elapsed times */
	end.tv_sec=(tv2.tv_sec-tv.tv_sec)-1;
	end.tv_usec=(1000000+tv2.tv_usec)-tv.tv_usec;
	end.tv_sec+=end.tv_usec/1000000;
	end.tv_usec%=1000000;

	mesg(M_DEBUG,"ELAPSED "
#ifdef HAVE_GETRUSAGE
		"CPU"
#endif
		": %lu.%.6lu", end.tv_sec, end.tv_usec);
#ifdef PROFILE_MODE
	pause();
#endif
}
