/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* TODO:
*  o Allow for specifying output file for elog concatenation
*  o Check for conflicting arguments
*/

#include <firestorm.h>
#include <dotscara/fdctl.h>

#include <sys/stat.h>
#include <sys/socket.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <fcntl.h>
#include <netinet/in.h>

#include <dotscara/mpool.h>
#include <dotscara/fobuf.h>
#include <loader.h>
#include <alert.h>
#include <args.h>
#include <target.h>
#include <signature.h>
#include <decode.h>
#include <detect.h>
#include <capture.h>
#include <mesg.h>
#include <elog.h>
#include <elog_read.h>
#include <index.h>
#include <cat.h>

/* Indexing library */
#include <idx.h>
#include <q_idx.h>
#include <i_idx.h>
#include <q_filter.h>

#include <firecat.h>

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"plugin-dir", 	1, 0, 'p'},
	{"options", 	1, 0, 'o'},
	{"format", 	1, 0, 'f'},
	{"index", 	0, 0, 'i'},
	{"list",	0, 0, 'l'},
	{"query", 	1, 0, 'q'},
	{"help", 	0, 0, 'h'},
	{NULL, 0, 0, 0}
};
#endif

/* Store argv[0] */
static char *cmd;

static int verbose_flag;

/* for MODE_CAT */
static char *oformat;
static char *p_options;

/* for MODE_IDX */
static char *query;
static void *filter;

/* Only print wanings,errors,and critical messages */
static void mesg_backend(unsigned char code, const char *str, size_t len)
{
	if ( !verbose_flag && code > M_WARN )
		return;

	/* Use stderr cos output may use stdout */
	fprintf(stderr, "%s: %s\n", cmd, str);
}

static void get_plugins(void)
{
	/* Load target plugins */
#if 1
	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");
#else
	loader_load_dir("./decode_plugins/.libs");
	loader_load_dir("./target_plugins/.libs");
#endif

	decode_load();
	target_load();
}

/* Display usage message and quit */
static void usage(int ecode) __noreturn;
static void usage(int ecode)
{
	extern struct target *targets;
	struct target *t;

	fprintf(stderr, "Usage: %s [-f format] [-o \"options\"] [FILE]...\n", cmd);
	fprintf(stderr, "Concatenates and converts firestorm extended log (elog) files.\n\n");
	fprintf(stderr, "    --format, -f      Choose a coversion format\n");
	fprintf(stderr, "    --options, -o     Plugin specific option string\n");
	fprintf(stderr, "    --plugin-dir, -p  Load extra plugins from this directory\n");
	fprintf(stderr, "    --index, -i       Create query indexes\n");
	fprintf(stderr, "    --list, -l        View index information for a file\n");
	fprintf(stderr, "    --query, -q       Select only certain packets\n");
	fprintf(stderr, "    --verbose, -v     Display debugging messages to stderr\n");
	fprintf(stderr, "    --help, -h        This message\n\n");
	fprintf(stderr, "With no FILE, or when FILE is -, read standard input.\n\n");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/protocols");
	fprintf(stderr, "Loading plugins from: %s\n\n", PLUGIN_DIR "/output");
	fprintf(stderr, "Available conversion formats and their options:\n");

	get_plugins();

	/* Display available target plugins */
	for(t=targets; t; t=t->next) {
		struct arg *a;

		fprintf(stderr, "  %s - %s\n", t->name, t->desc);
		if ( t->args == NULL )
			continue;

		for(a = t->args; a->name; a++)
			fprintf(stderr, "   o %s: %s (%s)\n",
				a->name,
				a->help ? a->help : "[no help available]",
				args_typename(a->type));
	}

	fprintf(stderr, "\nReport bugs to <gianni@scaramanga.co.uk>\n");
	firestorm_exit(ecode,NULL);
}

static int cat_file(int fd, const char *fn)
{
	return cat_elog(fd, fn, filter);
}

static void cat_init(void)
{
	get_plugins();

	if ( !cat_prepare_query(query, &filter) )
		firestorm_exit(EXIT_FAILURE, NULL);

	if ( !cat_prepare(oformat, p_options) )
		firestorm_exit(EXIT_FAILURE, NULL);
}

static void cat_fini(void)
{
	cat_free_query(filter);

	if ( !cat_close() )
		firestorm_exit(EXIT_FAILURE, "cat_close(): %s",
				sys_err2(NULL));
}

/* FIXME: Memory leaks */
static int lst_file(int fd, const char *fn)
{
	size_t fnlen = strlen(fn);
	char fnbuf[fnlen + 5];
	struct q_db *db;
	struct stat st1, st2;
	const char *efn=NULL;
	int efd = -1;
	int ood = 0;
	unsigned int i;

	/* See if we have an index only, or an elog plus an index */
	if ( fnlen > 4 && strncmp(fn + fnlen - 4, ".idx", 4) ) {
		snprintf(fnbuf, sizeof(fnbuf), "%s.idx", fn);
		efn = fn;
		efd = fd;
		fn = fnbuf;

		if ( fstat(fd, &st1) ) {
			mesg(M_ERR, "%s: fsync(): %s", efn, sys_err());
			return 0;
		}

		fd = open(fn, O_RDONLY);
		if ( fd < 0 ) {
			mesg(M_ERR, "%s: open(): %s", fn, sys_err());
			return 0;
		}

		if ( fstat(fd, &st2) ) {
			mesg(M_ERR, "%s: fsync(): %s", fn, sys_err());
			return 0;
		}

		if ( st1.st_mtime >= st2.st_mtime )
			ood = 1;
	}

	/* Print out filenames */
	db = q_idx_open(fd);
	if ( db == NULL ) {
		mesg(M_ERR, "%s: q_idx_open(): %s", fn, sys_err());
		return 0;
	}

	printf("idx: %s\n", fn);
	if ( efn ) {
		printf("elog: %s\n", efn);
		printf("status: %s\n", ood ? "out of date" : "up to date");
	}

	/* Print all of the fields */
	for(i=0; i < db->num_idx; i++) {
		const char *type;

		switch(db->idx[i].type) {
		case IDX_ORDERED_UINT32:
			type = "u_int32_t";
			break;
		case IDX_SKUNK:
			type = "string";
			break;
		case IDX_SKUNK_NOCASE:
			type = "string (nocase)";
			break;
		default:
			type = "invalid";
			break;
		}

		printf("field: name=%s num_keys=%u num_vals=%u type=%s\n",
				db->idx[i].name,
				db->idx[i].num_keys,
				db->idx[i].num_vals,
				type);
	}

	printf("\n");

	/* Close the files */
	fd_close(fd);
	if ( efn )
		fd_close(efd);
	return 1;
}

/* =====[ Index File Creation Interface ]===== */
static void idx_init(void)
{
	get_plugins();
}

static void idx_fini(void)
{
}

/* Define the various modes of operation */
static struct firecat_mode mode[]={
	[ MODE_CAT ] { .init=cat_init, .file=cat_file, .fini=cat_fini },
	[ MODE_QER ] { .init=cat_init, .file=cat_file, .fini=cat_fini },
	[ MODE_IDX ] { .init=idx_init, .file=index_create, .fini=idx_fini },
	[ MODE_LST ] { .init=NULL, .file=lst_file, .fini=NULL },
};

int main(int argc, char **argv)
{
	int m=MODE_CAT;
	int c, opti;
	int fd;
	int i;
	int ret = 0;

#ifdef HAVE_TZSET
	/* Set timezone from locale */
	tzset();
#endif

	/* Initialise what we need to */
	cmd=argv[0];

	mesg_set_backend(mesg_backend);

	firestorm_init();

	/* Do command line args */
#ifdef HAVE_GETOPT_LONG
	while( (c=getopt_long(argc, argv, "p:q:if:lo:vh?",
		opts, &opti))!=-1 )
#else
	while( (c=getopt(argc, argv, "p:q:if:lo:vh?"))!=-1 )
#endif
	{
		switch (c) {
		case 'i':
			m = MODE_IDX;
			break;
		case 'l':
			m = MODE_LST;
			break;
		case 'q':
			m = MODE_QER;
			query = optarg;
			break;
		case 'p':
			loader_load_dir(optarg);
			break;
		case 'f':
			if ( oformat )
				usage(EXIT_FAILURE);
			oformat=optarg;
			break;
		case 'o':
			if ( p_options )
				usage(EXIT_FAILURE);
			p_options=optarg;
			break;
		case 'v':
			verbose_flag = 1;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case '?':
		default:
			usage(EXIT_FAILURE);
		};
	}

	if ( mode[m].init )
		mode[m].init();

	/* cat semantics. if no file specified, or file is -, read
	 * standard input */
	if ( optind >= argc ) {
		if ( !mode[m].file(STDIN_FILENO, "(stdin)") )
			ret = 1;
	}

	for(i=optind; i<argc; i++ ) {
		/* use stdin for - */
		if ( !strcmp(argv[i],"-") ) {
			if ( !mode[m].file(STDIN_FILENO, "(stdin)") )
				ret = 1;
		}else{
			/* Open the input file */
			if ( (fd=open(argv[i],O_RDONLY))<0 ) {
				mesg(M_ERR, "%s: open(): %s",
					argv[i], sys_err());
				continue;
			}

			if ( !mode[m].file(fd, argv[i]) )
				ret = 1;
			fd_close(fd);
		}
	}

	if ( mode[m].fini )
		mode[m].fini();

	if ( ret ) {
		firestorm_exit(EXIT_FAILURE, "At least one operation failed");
	}else{
		firestorm_exit(EXIT_SUCCESS, NULL);
	}
}
