/*
* This file is part of dotscara
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/poll.h>

#include <dotscara/dotscara.h>
#include <dotscara/fdctl.h>
#include <dotscara/flbuf.h>

#define DEFAULT_BUF_SZ	4096
#define INCR_SZ		1024

static int ok_cbfn(void *priv, u_int8_t *buf, size_t line, size_t len)
{
	return 1;
}

static int err_cbfn(void *priv, u_int8_t *buf, size_t line, size_t len)
{
	return 0;
}

int __public flbuf_attach(struct flbuf *b, int fd, size_t max_sz,
				flbuf_cbfn_t cbfn, flbuf_cbfn_t cbfn2,
				void *priv)
{
	if ( cbfn == NULL )
		cbfn = ok_cbfn;
	if ( cbfn2 == NULL )
		cbfn2 = err_cbfn;

	b->fd = fd;
	b->max_sz = max_sz;
	b->cbfn = cbfn;
	b->cbfn2 = cbfn2;
	b->priv = priv;
	b->buf_sz = DEFAULT_BUF_SZ;
	b->buf = malloc(b->buf_sz);
	if ( b->buf == NULL )
		return 0;

	return 1;
}

void flbuf_detach(struct flbuf *b)
{
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->cbfn = NULL;
	b->priv = NULL;
	b->buf_sz = 0;
	b->max_sz = 0;
}

void flbuf_close(struct flbuf *b)
{
	if ( b->fd == -1 )
		goto noclose;

	__fd_close(b->fd);

noclose:
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->cbfn = NULL;
	b->priv = NULL;
	b->buf_sz = 0;
	b->max_sz = 0;
	b->fd = -1;
}

void flbuf_newfd(struct flbuf *b, int fd)
{
	if ( b->fd == -1 )
		goto noclose;

	__fd_close(b->fd);

noclose:
	b->fd = fd;
}

static inline size_t readline(u_int8_t *buf, size_t len, size_t *llen)
{
	u_int8_t *cur, *end;
	size_t ret;
	static int lastr=0;

	if ( lastr == 1 && len ) {
		if ( *buf == '\n' ) {
			buf++;
			len--;
		}
		lastr = 0;
	}

	for(cur=buf, end=buf+len, ret=0; cur < end; cur++) {
		ret++;

		switch ( *cur ) {
		case '\r':
			*cur = '\0';
			*llen = ret - 1;
			if ( ++cur < end ) {
				if ( *cur == '\n' )
					ret++;
			}else{
				lastr = 1;
			}
			return ret;
		case '\n':
			*cur = '\0';
			*llen = ret - 1;
			return ret;
		default:
			break;
		}
	}
	return 0;
}

int __public flbuf_read(struct flbuf *b)
{
	int eof = 0;
	size_t valid;
	size_t llen, len, line = 0;
	size_t residual = 0;
	u_int8_t *buf;
	int longline = 0;

	errno = 0;

	buf = b->buf;
	valid = b->buf_sz;

	while ( !eof ) {
		if ( !__fd_read(b->fd, buf, &valid, &eof) )
			return 0;

		buf = b->buf;
		valid += residual;

		while ( (llen = readline(buf, valid, &len)) ) {
			line++;
			if ( longline ) {
				longline = 0;
			}else if ( len != 0 ) {
				if ( !b->cbfn(b->priv, buf, line, len) )
					return 0;
			}
			buf += llen;
			valid -= llen;
		}

		if ( valid ) {
			size_t new_sz;
			u_int8_t *new;

			if ( valid != b->buf_sz ) {
				residual = valid;
				memmove(b->buf, buf, residual);
				buf = b->buf + residual;
				valid = b->buf_sz - residual;
				continue;
			}


			/* Line is too long */
			if ( b->max_sz != 0 && b->buf_sz >= b->max_sz ) {
				if ( !b->cbfn2(b->priv, NULL, line+1, 0) )
					return 0;
				longline = 1;
				residual = 0;
				buf = b->buf;
				valid = b->buf_sz;
				continue;
			}

			/* Grow the buffer and go again */
			new = NULL;
			new_sz = b->buf_sz + INCR_SZ;
			if ( b->max_sz && new_sz > b->max_sz )
				new_sz = b->max_sz;

			new = realloc(b->buf, new_sz);
			if ( new == NULL )
				return 0;

			buf = b->buf + b->buf_sz;
			residual = b->buf_sz;
			valid = (new_sz - b->buf_sz);

			b->buf = new;
			b->buf_sz = new_sz;
			continue;
		}else{
			residual = 0;
			buf = b->buf;
			valid = b->buf_sz;
		}
	}

	return 1;
}
