/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
 *
 * TODO:
 *  o Remove spooling code from core and import
 *  o Allow spool group configuration of target
 *  o Make firestorm-nids targets configurable
 *  o Allow source != FPS_ELOG for the target
 *  o When not doing buffering don't use fobuf, use writev
 *  o Re-write documentation for logging
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <target.h>
#include <plugin.h>
#include <capture.h>
#include <elog.h>
#include <elog_read.h>

static int elog_alert(struct packet *p, void *priv)
{
	struct fobuf *f = priv;
	char *pkt;
	size_t len;

	switch ( p->source ) {
	case FPS_ELOG:
		pkt = p->u.elog->file->map + p->u.elog->ofs;
		len = ntohl(p->u.elog->hdr->h.reclen);

		if ( !fobuf_write(f, pkt, len) ) {
			mesg(M_ERR, "fobuf_write(): %s", sys_err2(NULL));
			return 0;
		}
		break;
	default:
		return 0;
	}

	return 1;
}

static struct arg elog_args[]={
	{"file", ARGTYPE_FOBUF, .help = "Output file"},
	{NULL, ARGTYPE_NOP, NULL}
};

static int elog_validate(const char *args, void **ptr)
{
	struct efile_hdr fhdr;
	struct fobuf *f;

	f = calloc(1, sizeof(*f));
	if ( f == NULL ) {
		mesg(M_ERR, "calloc(): %s", sys_err());
		goto err;
	}

	f->fd = -1;
	elog_args[0].val.fobuf = f;

	switch( args_parse(elog_args, args, NULL) ) {
	case -1:
		mesg(M_ERR, "elog: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	if ( f->fd == -1 ) {
		if ( !fobuf_attach(f, STDOUT_FILENO, 0, 0) ) {
			mesg(M_ERR, "fobuf_attach(): %s", sys_err2(NULL));
			goto err;
		}
	}

	fhdr.magic = htonl(EF_MAGIC);
	fhdr.flags = 0;
	fhdr.vers_minor = EF_VERS_MIN;
	fhdr.vers_major = EF_VERS_MAJ;
	if ( !fobuf_write(f, &fhdr, sizeof(fhdr)) ) {
		mesg(M_ERR, "fobuf_write() header: %s", sys_err2(NULL));
		goto err;
	}

	*ptr = f;
	return 1;

err:
	if ( f ) {
		if ( f->fd != -1 )
			fobuf_close(f);
		free(f);
	}
	return 0;
}

static int elog_cleanup(void *priv)
{
	struct fobuf *f = priv;
	int ret = 1;

	if ( !fobuf_close(f) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(f);

	return ret;
}

struct target firestorm_targets[]={
	{
		.name = "elog",
		.desc = "Firestorm extended log (elog) files.",
		.target = elog_alert,
		.validate = elog_validate,
		.cleanup = elog_cleanup,
		.args = elog_args,
	},
	{.name = NULL,}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "target.elog",
	.desc		= "Firestorm exetended logs (elog)",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 0,
	.ver_minor	= 1,
};
